#include "stdafx.h"
#include "structs.h"
#include "WinstaDesk.h"
#include "DesktopHeapInfo.h"
#include "structs.h"
#include <cwchar> // wmemchr
#include <limits> // std::numeric_limits
#include <malloc.h> // alloca
#include "list.h"
#include "Process.h"
#include "Undeclared.h"

#pragma intrinsic(_bittestandset)

static const USHORT wMaximumIntBuffer = std::numeric_limits<ULONG>::digits10 + 1;

NTSTATUS ValidateHWinSta(HANDLE hWinSta, PWINDOWSTATION* ppWinSta)
{
	PAGED_CODE();
	ASSERT(ppWinSta && "Invalid NULL parameters");
	//
	// the original driver used NULL as the third parameter
	// we use the real type to enforce a type check
	//
	NTSTATUS stat = ObReferenceObjectByHandle(
		hWinSta, 
		STANDARD_RIGHTS_READ,
		*ExWindowStationObjectType,
		KernelMode,
		reinterpret_cast<PVOID*>(ppWinSta),
		NULL);
	if(NT_SUCCESS(stat))
	{
		ObDereferenceObject(*ppWinSta);
	}
	return stat;
}

NTSTATUS ValidateObject(PVOID pObject, POBJECT_TYPE pObjectType)
{
	PAGED_CODE();
	return ObReferenceObjectByPointer(pObject, STANDARD_RIGHTS_READ, pObjectType, KernelMode);
}

NTSTATUS GetObjectName(PVOID pObject, PUNICODE_STRING pusName)
{
	ASSERT(pusName && "Unicode string buffer is required");
	PAGED_CODE();
	NTSTATUS stat;
	UCHAR oniBuffer[1024];
	ULONG dwRealLenRequired;
	POBJECT_NAME_INFORMATION poni = reinterpret_cast<POBJECT_NAME_INFORMATION>(oniBuffer);
	if(NT_SUCCESS(stat = ObQueryNameString(pObject, poni, sizeof(oniBuffer), &dwRealLenRequired)))
	{
		// trim off everything but the actual object name
		// as that is what dheapmon is set to output
		// (the full directory path of is too long to display in its nicely layed out table)
		USHORT wBufferCharLength = poni->Name.Length / sizeof(WCHAR);
		WCHAR* pLastSlashChar = poni->Name.Buffer;
		WCHAR* pNextSlashChar = pLastSlashChar;
		// find the last slash
		// not using wcsrchr since the name string isn't necessarily null terminated
		while((pNextSlashChar = wmemchr(pNextSlashChar, L'\\', wBufferCharLength)) != NULL)
		{
			++pNextSlashChar;
			wBufferCharLength -= static_cast<USHORT>(pNextSlashChar - pLastSlashChar);
			pLastSlashChar = pNextSlashChar;
		}
		USHORT wBytesLeftInName = wBufferCharLength * sizeof(WCHAR);
		UNICODE_STRING usNameStringFromLastSlash = {wBytesLeftInName, wBytesLeftInName, pLastSlashChar};
		RtlCopyUnicodeString(pusName, &usNameStringFromLastSlash);
	}
	return stat;
}

PHEAP GetPHeapFromPDesktop(PDESKTOP pDesk)
{
	ASSERT(pDesk && "Valid pDesk is required");
	PAGED_CODE();
	switch(g_dwMajorVersion)
	{
		case 6:
		{
			switch(g_dwMinorVersion)
			{
				case 0:
				{
					// on vista and server 2008, it is the tagWIN32HEAP structure
					return pDesk->pHeap;
				}
				break;
				case 1:
				{
					// on 7, the heap is in the tagWIN32HEAP structure
					return pDesk->pheapDesktop->Heap;
				}
				break;
			}
		}
		break;
	}
	// this should never happen
	DBG_TEXT(
		"Windows version was validated at 6.0 or 6.1, "
		"but yet we seem to think it is %d.%d",
		g_dwMajorVersion,
		g_dwMinorVersion
		);
	return NULL;
}

NTSTATUS CreateDesktopEntry(PDESKTOP pDesk)
{
	ASSERT(pDesk && "Invalid desktop passed");
	PAGED_CODE();
	NTSTATUS stat = STATUS_SUCCESS;
	// validate the heap first
	PHEAP pWin32Heap = GetPHeapFromPDesktop(pDesk);
	if((pWin32Heap) && (pWin32Heap->Signature == g_dwHeapSignature))
	{
		PWINDOWSTATION pWinStaParent = pDesk->rpwinstaParent;
		DesktopHeapInfoEntry* pDeskHeapEntry = new DesktopHeapInfoEntry();
		if(!pDeskHeapEntry)
		{
			DBG_TEXT("No memory for DesktopHeapInfoEntry");
			return STATUS_INSUFFICIENT_RESOURCES;
		}
		// for easier typing
		DesktopHeapInfo* pDeskHeapInfo = &(pDeskHeapEntry->dhi);
		// fill in the windowstation name
		UNICODE_STRING usWinStaName = {0, sizeof(pDeskHeapInfo->wszWinStaName) - sizeof(WCHAR), pDeskHeapInfo->wszWinStaName};
		if(!NT_SUCCESS(stat = GetObjectName(pWinStaParent, &usWinStaName)))
		{
			DBG_TEXT("Failed to get windowstation name for object %p", pWinStaParent);
			delete pDeskHeapEntry;
			return stat;
		}
		DBG_TEXT_U(1, ("%p - Found windowstation ", pWinStaParent), &usWinStaName);
		// then the desk name
		UNICODE_STRING usDeskName = {0, sizeof(pDeskHeapInfo->wszDeskName) - sizeof(WCHAR), pDeskHeapInfo->wszDeskName};
		if(!NT_SUCCESS(stat = GetObjectName(pDesk, &usDeskName)))
		{
			DBG_TEXT("Failed to get desktop name for object %p", pDesk);
			delete pDeskHeapEntry;
			return stat;
		}
		DBG_TEXT_U(1, ("%p - Found desktop ", pDesk), &usDeskName);
		// because we can now get information about all sessions
		// we append the session id on to the desk name so the user can distinguish
		// them. Otherwise it looks like we report the same one multiple times
		// since there's usually a WinSta0 in every session
		//
		// first convert the session id to a unicode string
		WCHAR wszSessionId[wMaximumIntBuffer] = {0};
		UNICODE_STRING usSessionId = {0, wMaximumIntBuffer * sizeof(WCHAR), wszSessionId};
		RtlIntegerToUnicodeString(pDesk->dwSessionId, 10, &usSessionId);
		// then append on various bits to the desktop name
		RtlAppendUnicodeToString(&usDeskName, L" (s");
		RtlAppendUnicodeStringToString(&usDeskName, &usSessionId);
		RtlAppendUnicodeToString(&usDeskName, L")");
		// then fill in the rest of the parameters
		pDeskHeapInfo->dwSessionId = pWinStaParent->dwSessionId;
		pDeskHeapInfo->pWinSta = pWinStaParent;
		pDeskHeapInfo->pDesk = pDesk;
		pDeskHeapInfo->llHeapDesktop = reinterpret_cast<LONGLONG>(pWin32Heap);
		pDeskHeapInfo->dwTotalFreeBytes = pWin32Heap->TotalFreeSize * 8;
		pDeskHeapInfo->dwNumberOfPages = pWin32Heap->NumberOfPages;
		pDeskHeapInfo->dwNumberOfUncommittedPages = pWin32Heap->NumberOfUncommittedPages;
		// finally add the entry to our global list, which will hopefull be read later
		PushEntryList(&g_DesktopList, &(pDeskHeapEntry->nextEntry));
	}
	else
	{
		DBG_TEXT("Desktop %p didn't have a heap, or there was a signature mismatch", pDesk);
		stat = STATUS_ACCESS_DENIED;
	}
	return stat;
}

NTSTATUS EnumDesktops(PDESKTOP pDesk)
{
	PAGED_CODE();
	// cycle through the desktops
	// creating an entry in the global list for each one
	while(pDesk)
	{
		NTSTATUS stat = STATUS_SUCCESS;
		if(NT_SUCCESS(stat = ValidateObject(pDesk, *ExDesktopObjectType)))
		{
			stat = CreateDesktopEntry(pDesk);
			ObDereferenceObject(pDesk);
		}
		else
		{
			DBG_TEXT("Failed to validate desktop %p because of error 0x%x", pDesk, stat);
		}
		if(!NT_SUCCESS(stat))
		{
			return stat;
		}
		pDesk = pDesk->rpDeskNext;
	}
	DBG_TEXT("End of desk enumeration");
	return STATUS_SUCCESS;
}

NTSTATUS EnumWindowStations(PWINDOWSTATION pWinSta)
{
	PAGED_CODE();
	// this is of little benefit to the object we validated to start the process 
	// but might just stop any of the linked windowstations from
	// being deleted while we're working on them
	while(pWinSta)
	{
		NTSTATUS stat = STATUS_SUCCESS;
		if(NT_SUCCESS(stat = ValidateObject(pWinSta, *ExWindowStationObjectType)))
		{
			stat = EnumDesktops(pWinSta->rpDeskList);
			ObDereferenceObject(pWinSta);
		}
		else
		{
			DBG_TEXT("Failed to validate windowstation %p because of error 0x%x", pWinSta, stat);
		}
		if(!NT_SUCCESS(stat))
		{
			return stat;
		}
		pWinSta = pWinSta->rpWinstaNext;
	}
	DBG_TEXT("End of winsta enumeration");
	return STATUS_SUCCESS;
}

NTSTATUS GetPrimaryWinsta(PUNICODE_STRING pusDirectory, PVOID* ppWinstaObject)
{
	ASSERT(pusDirectory && ppWinstaObject && "Invalid arguments");
	PAGED_CODE();
	// try and open the directory
	OBJECT_ATTRIBUTES oa = RTL_CONSTANT_OBJECT_ATTRIBUTES(pusDirectory, OBJ_CASE_INSENSITIVE | OBJ_KERNEL_HANDLE);
	HANDLE hDirectory = NULL;
	NTSTATUS stat = ZwOpenDirectoryObject(&hDirectory, DIRECTORY_QUERY, &oa);
	DBG_TEXT("ZwOpenDirectoryObject (%wZ) returned 0x%x", pusDirectory, stat);
	if(!NT_SUCCESS(stat))
	{
		return stat;
	}
	WinType<> dirWrap(hDirectory, &ZwClose);
	UNICODE_STRING usWinsta0 = RTL_CONSTANT_STRING(L"WinSta0");
	// reuse the object attributes
	oa.RootDirectory = hDirectory;
	oa.ObjectName = &usWinsta0;
	HANDLE hWinsta = NULL;
	// try and get a handle to winsta0 in the directory
	stat = ObOpenObjectByName(&oa, *ExWindowStationObjectType, KernelMode, NULL, STANDARD_RIGHTS_READ, NULL, &hWinsta);
	if(!NT_SUCCESS(stat))
	{
		DBG_TEXT("ObOpenObjectByName failed, error 0x%x", stat);
		return stat;
	}
	// turn the handle to a pointer
	stat = ObReferenceObjectByHandle(hWinsta, STANDARD_RIGHTS_READ, *ExWindowStationObjectType, KernelMode, ppWinstaObject, NULL);
	ZwClose(hWinsta);
	return stat;
}

NTSTATUS GetSessionPrimaryWinsta(ULONG dwSessionId, PVOID* ppWinsta)
{
	ASSERT(ppWinsta && "Invalid output pointer given");
	PAGED_CODE();
	NTSTATUS stat = STATUS_SUCCESS;
	// valid windowstations are recorded in the
	// \Sessions\<session id>\Windows\WindowStations directory
	// ones in session 0 may also be kept in the \Windows\WindowStations global directory
	static const WCHAR wszSessionPrefix[] = L"\\Sessions\\";
	static const WCHAR wszSessionSuffix[] = L"\\Windows\\WindowStations";
	static const USHORT wPrefixSuffixLength = ARRAYSIZE(wszSessionPrefix) + ARRAYSIZE(wszSessionSuffix);;

	// convert session id to string
	WCHAR wszIntBuffer[wMaximumIntBuffer];
	UNICODE_STRING usIntBuffer = {0, (wMaximumIntBuffer * sizeof(WCHAR)), wszIntBuffer};
	RtlIntegerToUnicodeString(dwSessionId, 10, &usIntBuffer);

	USHORT wTotalStringLength = wPrefixSuffixLength + usIntBuffer.Length + 1;
	ScopedArray<WCHAR> saSessionDirBuffer(new WCHAR[wTotalStringLength]);
	if(*saSessionDirBuffer)
	{
		// concat everything together to form the full directory path
		UNICODE_STRING usSessionDir = {0, wTotalStringLength * sizeof(WCHAR), *saSessionDirBuffer};
		RtlAppendUnicodeToString(&usSessionDir, wszSessionPrefix);
		RtlAppendUnicodeStringToString(&usSessionDir, &usIntBuffer);
		RtlAppendUnicodeToString(&usSessionDir, wszSessionSuffix);
		// try and get the first winsta
		if(!NT_SUCCESS(stat = GetPrimaryWinsta(&usSessionDir, ppWinsta)))
		{
			// if we failed, and this is session 0
			// try the global version of the directory
			if(dwSessionId == 0)
			{
				usSessionDir.Length = 0;
				RtlAppendUnicodeToString(&usSessionDir, wszSessionSuffix);
				stat = GetPrimaryWinsta(&usSessionDir, ppWinsta);
			}
		}
	}
	return stat;
}

VOID EnumSessionsAndPrimaryWinstas(PDEVICE_OBJECT /*pDev*/, PVOID pContext)
{
	PAGED_CODE();
	WorkItemContext* pWorkCtx = static_cast<WorkItemContext*>(pContext);
	ASSERT(pWorkCtx && pWorkCtx->pExitEvent && pWorkCtx->pSysProcessBuffer && "Invalid context for work item");
	
	ULONG dwSessionsSeen[2] = {0};
	ULONG dwTotalBufferOffset = 0;
	UCHAR* pSysBuffer = pWorkCtx->pSysProcessBuffer;
	SYSTEM_PROCESS_INFORMATION* pSysProcess = reinterpret_cast<SYSTEM_PROCESS_INFORMATION*>(pSysBuffer);
	NTSTATUS stat = STATUS_SUCCESS;
	// loop through the processes we found, looking for unique sessions
	while(NT_SUCCESS(stat) && pSysProcess->NextEntryOffset)
	{
		PEPROCESS pEProcess = NULL;
		if(NT_SUCCESS(PsLookupProcessByProcessId(pSysProcess->UniqueProcessId, &pEProcess)))
		{
			ULONG dwSession = PsGetProcessSessionIdEx(pEProcess);
			if(dwSession != static_cast<ULONG>(-1))
			{
				// check it's within recordable range
				// the limit is currently 64 sessions which should be enough 
				// for everthing but the most terminal of services
				if(dwSession < (CHAR_BIT * sizeof(dwSessionsSeen)))
				{
					// see if we've already processed this session
					ULONG dwWhichPart = (dwSession >= (CHAR_BIT * sizeof(*dwSessionsSeen)));
					BOOLEAN bSeenSession = _bittestandset(reinterpret_cast<LONG*>(&dwSessionsSeen[dwWhichPart]), static_cast<LONG>(dwSession));
					// if we haven't try and get it's primary windowstation
					if(!bSeenSession)
					{
						DBG_TEXT("Found process %p in currently unenumerated session %lu", pEProcess, dwSession);
						PVOID pWinsta = NULL;
						stat = GetSessionPrimaryWinsta(dwSession, &pWinsta);
						// if we haven't got a winsta object, go to the next one
						// as this session might not have any windowstations
						if(pWinsta)
						{
							// Now attach this thread to the target process.
							// this is done as all the data relating to the windowstations
							// and desktops are in session space
							// which isn't readable from a different session.
							// attaching to a process in the target session ensures we can read it
							KAPC_STATE apcState = {0};
							DBG_TEXT("Attaching to process %p in session %d", pEProcess, dwSession);
							KeStackAttachProcess(pEProcess, &apcState);
							stat = EnumWindowStations(static_cast<PWINDOWSTATION>(pWinsta));
							KeUnstackDetachProcess(&apcState);
							ObDereferenceObject(pWinsta);
						}
						else
						{
							DBG_TEXT("Couldn't find primary windowstation for session %d", dwSession);
						}
					}
				}
				else
				{
					DBG_TEXT("Session %lu is active, but outside recordable range. Windowstations will be missed", dwSession);
				}
			}
			ObDereferenceObject(pEProcess);
		}
		dwTotalBufferOffset += pSysProcess->NextEntryOffset;
		pSysProcess = reinterpret_cast<SYSTEM_PROCESS_INFORMATION*>(&pSysBuffer[dwTotalBufferOffset]);
	}
	pWorkCtx->stat = stat;
	KeSetEvent(pWorkCtx->pExitEvent, 0, FALSE);
}

NTSTATUS TryEnumAllWindowStations(PDEVICE_OBJECT pDevObj)
{
	ASSERT(pDevObj && "Invalid device object passed");
	PAGED_CODE();
	NTSTATUS stat = STATUS_SUCCESS;
	WorkItemContext wic = {0};
	// if only this was a compile time constant
	ULONG dwWorkItemSize = IoSizeofWorkItem();
	PIO_WORKITEM pWorkItem = reinterpret_cast<PIO_WORKITEM>(_alloca(dwWorkItemSize));

	ScopedArray<UCHAR> saProcessesBuffer(NULL);
	// grab info on all the currently running processes
	if(!NT_SUCCESS(stat = GetRunningProcesses(&saProcessesBuffer)))
	{
		return stat;
	}
	wic.pSysProcessBuffer = *saProcessesBuffer;
	// we have to pass off the job to a work item / worker thread
	// here because the functions we need to use to get the windowstations
	// are documented as PASSIVE_LEVEL only, and we're at APC_LEVEL
	IoInitializeWorkItem(pDevObj, pWorkItem);
	KEVENT eWorkFinished;
	KeInitializeEvent(&eWorkFinished, NotificationEvent, FALSE);
	wic.pExitEvent = &eWorkFinished;
	IoQueueWorkItem(pWorkItem, &EnumSessionsAndPrimaryWinstas, DelayedWorkQueue, &wic);
	KeWaitForSingleObject(&eWorkFinished, Executive, KernelMode, FALSE, NULL); 
	IoUninitializeWorkItem(pWorkItem);
	DBG_TEXT("Work item returned 0x%x", wic.stat);
	stat = wic.stat;
	if(!NT_SUCCESS(stat))
	{
		// if we failed somewhere ensure the desk list is empty
		// otherwise there might be duplicates inserted by the old
		// enumeration method, or an incorrect number reported
		// or copied to dheapmon
		EmptyDeskList();
	}
	return stat;
}

NTSTATUS GetDesktopHeapInfo(HANDLE hWinSta, PDEVICE_OBJECT pDev)
{
	ASSERT(pDev && "Invalid device object passed");
	PAGED_CODE();
	// try and use the improved method which will scan all
	// windowstations & desktops regardless of session
	// over the old single-session only method implemented
	// by the original driver
	NTSTATUS stat = TryEnumAllWindowStations(pDev);
	if(!NT_SUCCESS(stat))
	{
		PWINDOWSTATION pWinsta;
		if(!NT_SUCCESS(stat = ValidateHWinSta(hWinSta, &pWinsta)))
		{
			DBG_TEXT("Failed to validate HWINSTA 0x%x. Error 0x%x", hWinSta, stat);
			return stat;
		}
		stat = EnumWindowStations(pWinsta);
	}
	return stat;
}